/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Nordic UART Bridge Service (NUS) sample
 */

#include <zephyr/types.h>
#include <zephyr.h>
#include <init.h>
#include <device.h>
#include <soc.h>
#include <power/power.h>
#include <drivers/gpio.h>
#include <drivers/i2c.h>
#include <drivers/adc.h>
#include <hal/nrf_saadc.h>
#include <random/rand32.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/uuid.h>
#include <bluetooth/gatt.h>
#include <bluetooth/hci.h>

#include <bluetooth/services/nus.h>
#include "homesmart.h"


#include <settings/settings.h>

#include <stdio.h>

#include <logging/log.h>

#define LOG_MODULE_NAME peripheral_uart
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define STACKSIZE CONFIG_BT_NUS_THREAD_STACK_SIZE
#define PRIORITY 7

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN	(sizeof(DEVICE_NAME) - 1)

#define RUN_STATUS_LED DK_LED1
#define RUN_LED_BLINK_INTERVAL 1000

#define CON_STATUS_LED DK_LED2

#define KEY_PASSKEY_ACCEPT DK_BTN1_MSK
#define KEY_PASSKEY_REJECT DK_BTN2_MSK

#define UART_BUF_SIZE CONFIG_BT_NUS_UART_BUFFER_SIZE
#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)
#define UART_WAIT_FOR_RX CONFIG_BT_NUS_UART_RX_WAIT_TIME
/* Note that the two hex variablesin the following define should set the advertising interval to between 2 and 2.4 seconds */
#define BT_LE_ADV_CONN_SLOW BT_LE_ADV_PARAM(BT_LE_ADV_OPT_CONNECTABLE, \
                                      0X0C80, \
                                       0X0F00, NULL)  
static K_SEM_DEFINE(ble_init_ok, 0, 1);

static struct bt_conn *current_conn;


struct bt_xmit_data {
	void *fifo_reserved;
	uint8_t data[UART_BUF_SIZE];
	uint16_t len;
};

static K_FIFO_DEFINE(fifo_bt_tx_data);

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};

static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NUS_VAL),
};
const struct device *gpio1;
const struct device *i2c0;
const struct device *adc0;

//struct k_timer msclock;
int millitime;
struct msgs *fstMsg=NULL, *lstMsg;
bool keypadactive = false;


extern void mstick(struct ktimer *msclock)
{millitime++;
}
K_TIMER_DEFINE(msclock, mstick,NULL);
/* debugging only     -   if messages were qued while there was no active blue tooth connection - we dump them to the newly connected device */
void dumpque()
{uint8_t queddata[100];
 struct bt_xmit_data *tx;
 uint32_t i,j,len1,len2;
 memset(queddata,0,sizeof(queddata));                                // clear the message work area
 strcpy(queddata,dequeMsg());                                        // and retrieve a message from the que
 i=strlen(queddata);
 if(queddata[i-1]=='\n')                                             // remove trailing new lines
    queddata[i-1]=0;
 while(queddata[0]!=0)                                               // while there is data left to send
     {
       tx = k_malloc(sizeof(*tx));                                   // allocate a buffer 
       if (!tx) {
	         LOG_WRN("Not able to allocate UART send data buffer");
 	         return;
	        }
       len1=strlen(queddata);                                         // get the size of data not yet sent
       i = sizeof(tx->data) - 1;                                      // get the max size of the buffer we're building to sent out
       if(len1<=i)                                                     // if there's enough room for the remaining data
           len2=len1;                                                 // we'll move all of it out
       else len2=i;                                                   // otherwise - we'll move as much as we can 
       memcpy(tx->data,queddata,len2);                                // move the selected data to the bt write buffer
       tx->len=len2;                                                  // ass well as the length
       k_fifo_put(&fifo_bt_tx_data, tx);                              // and send it on its way
       j=0;                                                           // now we set up to left shift the data we just sent our of the buffer
       for(i=len2;i<=sizeof(queddata);i++)                            // shift the buffer
           {
            queddata[j]=queddata[i];
            j++;
           }

     }

}
static void connected(struct bt_conn *conn, uint8_t err)
{       
	char addr[BT_ADDR_LE_STR_LEN];

	if (err) {
		LOG_ERR("Connection failed (err %u)", err);
		return;
	}

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
	LOG_INF("Connected %s", log_strdup(addr));

	current_conn = bt_conn_ref(conn);

}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Disconnected: %s (reason %u)", log_strdup(addr), reason);


	if (current_conn) {
		bt_conn_unref(current_conn);
		current_conn = NULL;
	//	dk_set_led_off(CON_STATUS_LED);
	}
}


static struct bt_conn_cb conn_callbacks = {
	.connected    = connected,
	.disconnected = disconnected,
};


static void bt_receive_cb(struct bt_conn *conn, const uint8_t *const data,
			  uint16_t len)
{
	char addr[BT_ADDR_LE_STR_LEN] = {0};

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, ARRAY_SIZE(addr));

	LOG_INF("Received data from: %s", log_strdup(addr));

	while(fstMsg!=NULL)                                            // if there are qued messages
           dumpque();
	for (uint16_t pos = 0; pos != len;) {
		struct bt_xmit_data *tx = k_malloc(sizeof(*tx));

		if (!tx) {
			LOG_WRN("Not able to allocate UART send data buffer");
			return;
		}

		/* Keep the last byte of TX buffer for potential LF char. */
		size_t tx_data_size = sizeof(tx->data) - 1;

		if ((len - pos) > tx_data_size) {
			tx->len = tx_data_size;
		} else {
			tx->len = (len - pos);
		}

		memcpy(tx->data, &data[pos], tx->len);

		pos += tx->len;

		/* Append the LF character when the CR character triggered
		 * transmission from the peer.
		 */
		if ((pos == len) && (data[len - 1] == '\r')) {
			tx->data[tx->len] = '\n';
			tx->len++;
		}
                // MKO CHANGED the following three lines of code, eliminating the whole uart construct from the program. It's now an echo server, simply sending 
                // what it received back to the sender. Tested on the NRF52 DK - seems to be no problems
		//err = uart_tx(uart, tx->data, tx->len, SYS_FOREVER_MS);
		//if (err)
                 {
			k_fifo_put(&fifo_bt_tx_data, tx);
		}
	}
}

static struct bt_nus_cb nus_cb = {
	.received = bt_receive_cb,
};

void error(void)
{
	//dk_set_leds_state(DK_ALL_LEDS_MSK, DK_NO_LEDS_MSK);

	while (true) {
		/* Spin for ever */
		k_sleep(K_MSEC(1000));
	}
}

static void configure_gpio(void)
{       gpio1=device_get_binding("GPIO_1");

        if (gpio1 == NULL) 
                printk("Error: didn't find gpio1  device\n");
        else printk("GPIO configured\n");
}

static void configure_i2c(void)
{       i2c0=device_get_binding("I2C_1");

        if (i2c0 == NULL) 
                printk("Error: didn't find i2c0  device\n");
         else printk("I2C configured\n");
}

static void configure_adc(void)
{       adc0=device_get_binding("ADC_0");

        if (adc0 == NULL) 
                printk("Error: didn't find i2c0  device\n");
        else printk("ADC configured\n");
}


void main(void)
{       bool ret;

	int err = 0;
        int i, loopct=0;
        bool fta=true;

	configure_gpio();
        //configure_i2c();
        //configure_adc();
        //k_timer_init(&msclock,mstick, NULL);
        ret=init_keys();
	bt_conn_cb_register(&conn_callbacks);

        i=sys_rand32_get();
        printk("RANDOM is %d\n",i);
	err = bt_enable(NULL);
	if (err) {
		error();
	}

	LOG_INF("Bluetooth initialized");

	k_sem_give(&ble_init_ok);

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	err = bt_nus_init(&nus_cb);
	if (err) {
		LOG_ERR("Failed to initialize UART service (err: %d)", err);
		return;
	}

	err = bt_le_adv_start(BT_LE_ADV_CONN_SLOW, ad, ARRAY_SIZE(ad), sd,
			      ARRAY_SIZE(sd));
	if (err) {
		LOG_ERR("Advertising failed to start (err %d)", err);
	}
        
	printk("Starting Nordic UART service example\n");

	for (;;) {k_sleep(K_FOREVER);
                   loopct++;
                   if((loopct%100000000)==0)
                       printk("LOOPCT %d\n",loopct);
                  if(keypadactive)
                   getkeypadinput();
                
                 continue;
	}
}

void ble_write_thread(void)
{
	/* Don't go any further until BLE is initialized */
	k_sem_take(&ble_init_ok, K_FOREVER);

	for (;;) {
		/* Wait indefinitely for data to be sent over bluetooth */
		struct bt_xmit_data *buf = k_fifo_get(&fifo_bt_tx_data,
						     K_FOREVER);

		if (bt_nus_send(NULL, buf->data, buf->len)) {
			LOG_WRN("Failed to send data over BLE connection");
		}

		k_free(buf);
	}
}

K_THREAD_DEFINE(ble_write_thread_id, STACKSIZE, ble_write_thread, NULL, NULL,
		NULL, PRIORITY, 0, 0);
